# Copyright 2025 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

from surfactant.utils.get_plugin_settings import extract_settings_from_docstring


def function_with_settings():
    """Function for testing: has a colon

    Config Options:
        test_function.option1 (str): Some option. [default="123"]
        test_function.option2 (int): Some other option. [default=123]
        test_function.option3 (int): Some long description that doesn't fit on
            a single line
        test_function.option4: Some option missing a type: and with a colon
    """


def test_extract_settings_from_docstring():
    assert function_with_settings.__doc__

    settings = extract_settings_from_docstring(function_with_settings.__doc__)
    assert len(settings) == 4

    assert settings[0].name == "test_function.option1"
    assert settings[0].type_ == "str"
    assert settings[0].description == "Some option."
    assert settings[0].default == '"123"'

    assert settings[1].name == "test_function.option2"
    assert settings[1].type_ == "int"
    assert settings[1].description == "Some other option."
    assert settings[1].default == "123"

    assert settings[2].name == "test_function.option3"
    assert settings[2].type_ == "int"
    assert settings[2].description == "Some long description that doesn't fit on a single line"
    assert settings[2].default is None

    assert settings[3].name == "test_function.option4"
    assert settings[3].type_ is None
    assert settings[3].description == "Some option missing a type: and with a colon"
    assert settings[3].default is None



import logging
from dataclasses import asdict, dataclass
from pathlib import Path
from typing import Any, Dict, List, Optional

import surfactant.plugin
from surfactant.sbomtypes import SBOM, Software

try:
    from dapper_python.databases.linuxDB import LinuxDB
    from dapper_python.dataset_loader import DatasetCatalog
    from dapper_python.normalize import NormalizedFileName, normalize_file_name

    DAPPER_AVAILABLE = True
except ImportError:
    DAPPER_AVAILABLE = False
    logging.warning("dapper-python not installed. Dapper plugin will be disabled.")


@dataclass
class DapperPackageInfo:
    """Information about a package found by Dapper."""

    package_name: str
    package_dataset: str
    original_name: str
    file_path: str
    normalized_name: Optional[str] = None
    version: Optional[str] = None
    soabi: Optional[str] = None

    @classmethod
    def from_result(cls, result, dataset_name, filename):
        """Create DapperPackageInfo from database query result."""

        normalized_result = normalize_file_name(filename)

        if isinstance(normalized_result, NormalizedFileName):
            version = normalized_result.version
            soabi = normalized_result.soabi
            normalized_name = normalized_result.name
        else:
            version = None
            soabi = None
            normalized_name = filename

        return cls(
            package_name=result.package_name,
            package_dataset=dataset_name,
            original_name=filename,
            normalized_name=normalized_name,
            file_path=str(result.file_path),
            version=version,
            soabi=soabi,
        )


class DapperPlugin:  # pylint: disable=too-few-public-methods
    """Plugin to integrate Dapper package detection with Surfactant."""

    def __init__(self):
        """Initialize the Dapper plugin with dataset catalog."""
        self.catalog = None
        self.linux_datasets = []
        self.nuget_dataset = None
        if DAPPER_AVAILABLE:
            self._init_catalog()

    def _init_catalog(self):
        """Initialize the dataset catalog and check available datasets."""
        try:
            # Try to load the dataset catalog
            self.catalog = DatasetCatalog(app_name="dapper")

            # Check for available datasets
            available = self.catalog.get_available_datasets()

            # Look for Linux and NuGet datasets
            for dataset in available:
                if any(distro in dataset for distro in ["debian", "ubuntu"]):
                    self.linux_datasets.append(dataset)

            if not self.linux_datasets:
                pass  # No Linux datasets found for Dapper plugin

        except Exception:  # pylint: disable=broad-exception-caught
            self.catalog = None

    def lookup_package(self, file_path: str, file_types: List[str]) -> Optional[Dict[str, Any]]:
        """Provides package lookup for files in SBOMs."""
        if not self.catalog or not DAPPER_AVAILABLE:
            return None

        # Extract just the filename from the path
        filename = Path(file_path).name

        # Determine file type from Surfactant's filetype list
        file_type = None
        for ft in file_types:
            if "ELF" in ft:
                file_type = "ELF"
                break
            if "PE" in ft or "PE32" in ft or "DLL" in ft:
                file_type = "PE"
                break

        # Also check by extension if not detected
        if not file_type:
            ext = Path(filename).suffix.lower()
            if ext in [".so", ".o"] or (not ext and "." not in filename):
                file_type = "ELF"
            elif ext in [".dll", ".exe", ".sys"]:
                file_type = "PE"

        if not file_type:
            return None

        all_results = []  # Initialize here!

        # Query all Linux datasets for ELF files
        if file_type == "ELF" and self.linux_datasets:  # Check list, not None variable
            for dataset_name in self.linux_datasets:
                try:
                    db_path = self.catalog.get_dataset_path(dataset_name)

                    # Path resolution
                    if db_path and not db_path.is_absolute():
                        app_dir = Path(self.catalog.get_app_data_dir(self.catalog.app_name))
                        db_path = app_dir / db_path

                    if not db_path or not db_path.exists():
                        continue

                    linux_db = LinuxDB(db_path)
                    results = linux_db.query_filename(filename, normalize=False)

                    if results:
                        # Add results from this dataset
                        packages = [
                            asdict(DapperPackageInfo.from_result(r, dataset_name, filename))
                            for r in results
                        ]
                        all_results.extend(packages)

                except Exception:  # pylint: disable=broad-exception-caught
                    pass  # Error querying dataset

        if all_results:
            return {"dapper_packages": all_results}

        return None


# Global plugin instance
_PLUGIN_INSTANCE = None


def _get_plugin():
    global _PLUGIN_INSTANCE  # pylint: disable=global-statement
    if _PLUGIN_INSTANCE is None:
        _PLUGIN_INSTANCE = DapperPlugin()
    return _PLUGIN_INSTANCE


@surfactant.plugin.hookimpl
def extract_file_info(
    sbom: SBOM,
    software: Software,
    filename: str,
    filetype: List[str],
) -> Optional[Dict[str, Any]]:
    if not DAPPER_AVAILABLE:
        return None

    plugin = _get_plugin()

    # Look up package information based on the file
    package_info = plugin.lookup_package(filename, filetype)

    if package_info:
        return package_info

    return None



# Copyright 2025 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

import re
from dataclasses import dataclass
from typing import Dict, List, Optional

from surfactant.plugin.manager import get_plugin_manager


@dataclass
class PluginSetting:
    name: str
    type_: Optional[str]
    description: str
    default: Optional[str]


def extract_plugin_settings() -> Dict[str, List[PluginSetting]]:
    """Extract settings information from plugin docstrings"""
    settings = {}
    pm = get_plugin_manager()
    for plugin in pm.get_plugins():
        if (
            plugin.__doc__
            and plugin.__doc__.find("Config Options:") != -1
            and plugin.settings_name()
        ):
            settings[plugin.settings_name()] = extract_settings_from_docstring(plugin.__doc__)
    return settings


__option_re = re.compile(r"([^(:]+)\s*([(][^)]+[)])?:(.*)")


def extract_settings_from_docstring(docstring: str) -> List[PluginSetting]:
    def count_leading_indentation(s: str) -> int:
        indent_amount = 0
        while indent_amount < len(s) and s[indent_amount].isspace():
            indent_amount += 1
        return indent_amount

    lines = docstring.splitlines()
    line_no = 0
    # Skip to the config options line
    while not lines[line_no].strip().startswith("Config Options:"):
        line_no += 1

    # Count the indentation of the next line (which should be the first option)
    line_no += 1
    indentation_amount = count_leading_indentation(lines[line_no])

    # Split by option (based on leading whitespace)
    settings_lines = []
    while line_no < len(lines) and count_leading_indentation(lines[line_no]) == indentation_amount:
        cur_line = lines[line_no].strip()
        while (
            line_no + 1 < len(lines)
            and count_leading_indentation(lines[line_no + 1]) > indentation_amount
        ):
            line_no += 1
            cur_line += " " + lines[line_no].strip()
        line_no += 1
        settings_lines.append(cur_line)

    settings = []
    for settings_line in settings_lines:
        m = __option_re.match(settings_line)
        if m:
            name, type_, desc = m.groups()
            # Remove the parentheses if a type was matched
            if type_:
                type_ = type_.strip()[1:][:-1].strip()
            default_value = None
            # Search for the start of the default string
            loc = desc.rfind("[default=")
            if loc != -1:
                end_loc = desc.rfind("]")
                default_value = desc[loc + 9 : end_loc].strip()
                desc = desc[:loc].strip()
            settings.append(PluginSetting(name.strip(), type_, desc.strip(), default_value))

    return settings



#!/usr/bin/env python3

import dataclasses
import os
import sys
from argparse import ArgumentParser
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import List

import networkx
import pyvis
from dataclasses_json import dataclass_json
from jinja2 import Environment, FileSystemLoader
from loguru import logger

from surfactant.sbomtypes import SBOM


@dataclass
class DisplaySettings:
    fg_color: str = "white"

    node_scale_factor: int = 4  # Increase node size by degree
    icon_scale_factor: int = 4

    container_max_size: int = 400


class NodeType(str, Enum):
    Unknown = ("Unknown",)
    File = ("File",)
    Container = "Container"


@dataclass_json
@dataclass
class NodeMetadata:
    type: NodeType = NodeType.File
    hidden: bool = False
    nodeFileName: str = ""
    SBOMFileName: str = ""


def generate_dependency_graph(
    sbomDict: SBOM,
    enableCulling: bool = False,
) -> networkx.graph:
    g = sbomDict["sbom"].graph

    # Add attributes to nodes
    for nodeID, attrib in g.nodes(data=True):
        entry = next((e for e in sbomDict["sbom"].software if e.UUID == nodeID), None)
        if entry is None:
            logger.error("Malformed SBOM, can't find Software entry with UUID: {nodeID}")
            sys.exit(-1)

        fileName = (entry.fileName or [entry.UUID])[0]
        metadata = NodeMetadata(
            nodeFileName=fileName,  # Label and file name can differ
            SBOMFileName=sbomDict["sbomFileName"],
        )

        attrib["label"] = fileName
        attrib["surfactantSoftwareStruct"] = dataclasses.asdict(entry)
        attrib["nodeMetadata"] = dataclasses.asdict(metadata)

    # Add attributes to edges
    for u, _, key, attrib in g.edges(keys=True, data=True):
        isContainer = key == "Contains"

        if isContainer:
            attrib["dashes"] = True
            g.nodes[u]["nodeMetadata"]["type"] = NodeType.Container

        attrib["container"] = isContainer

    g = g.reverse()  # Pyvis expects an inverted layout (edges inadvertently swap direction)

    if enableCulling:
        g.remove_nodes_from(list(networkx.isolates(g)))  # Cull nodes with no edges
    elif networkx.number_of_isolates(g) > 500:
        logger.info(
            "Large number of isolate nodes (nodes without any edges) detected. Consider re-running with --cull to improve performance"
        )

    # Set node size based on connected edges
    updatedValues = {}
    for ID in g.nodes:
        fileType = g.nodes[ID]["nodeMetadata"]["type"]
        if fileType == NodeType.File:
            updatedValues[ID] = DisplaySettings.icon_scale_factor * max(2, g.in_degree(ID))
        elif fileType == NodeType.Container:
            updatedValues[ID] = DisplaySettings.icon_scale_factor * max(1, g.degree(ID))
            if updatedValues[ID] > DisplaySettings.container_max_size:
                updatedValues[ID] = DisplaySettings.container_max_size
                logger.info(
                    f"Container with UUID: {ID} exceeds max width, constraining to {DisplaySettings.container_max_size} to ensure graph legibility"
                )

    networkx.set_node_attributes(g, updatedValues, "size")

    return g


def generate_pyvis_graph(
    graph: networkx.graph, filename: str, use_progress_bar: bool = False
) -> None:
    pv = pyvis.network.Network(
        directed=True,
        height="100vh",
        width="100vw",
        font_color=DisplaySettings.fg_color,
        cdn_resources="remote",
    )

    currentDir = os.path.dirname(os.path.abspath(__file__))
    pv.template_dir = os.path.join(
        currentDir, "./customTemplates/"
    )  # Use custom template to fix CSS issue with not filling page
    pv.templateEnv = Environment(loader=FileSystemLoader(pv.template_dir))

    pv.force_atlas_2based()

    pv.options.interaction.selectConnectedEdges = True

    if graph.number_of_nodes() > 600 and not use_progress_bar:
        logger.info(
            f"Large graph detected. Disabling physics to speed up loading time, this can be re-enabled by using the toggle on the left side of the page. Disable this behavior by re-running {Path(__file__).name} with -pb/--use-progress-bar"
        )
        pv.toggle_physics(False)

    pv.from_nx(graph)

    # Replace node shape with icon if a container
    for n in pv.nodes:
        if n["nodeMetadata"]["type"] == NodeType.Container:
            n["shape"] = "dot"  # Embed icon in shape
            n["icon"] = {
                "face": "'Font Awesome 6 Free'",
                "weight": "bold",
                "code": "\uf07c",  # Folder open icon
                "color": DisplaySettings.fg_color,
            }
        if n["nodeMetadata"]["type"] == NodeType.File:
            n["shape"] = "dot"
            n["icon"] = {
                "face": "'Font Awesome 6 Free'",
                "weight": "bold",
                "code": "\uf15b",  # fa-file
                "color": DisplaySettings.fg_color,
            }

    pv.write_html(filename, notebook=False)


def main():
    parser = ArgumentParser(description="Surfactant SBOM Visualization")
    parser.add_argument(
        "-p",
        "--path",
        nargs="+",
        help="Path(s) to JSON SBOMs",
        required=True,
    )

    parser.add_argument(
        "-c",
        "--cull",
        default=False,
        action="store_true",
        help="Enable culling of isolated nodes (may improve performance on large graphs at the cost of node completeness)",
    )

    parser.add_argument(
        "-pb",
        "--use-progress-bar",
        default=False,
        action="store_true",
        help="Display progress bar while waiting for large graphs to load instead of disabling physics",
    )

    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)

    args = parser.parse_args()

    sboms: List[dict] = []
    for path in args.path:
        with open(path, "r") as f:
            sboms.append({"sbom": SBOM.from_json(f.read()), "sbomFileName": Path(path).name})

    g = generate_dependency_graph(sboms[0], args.cull)

    filename = Path(args.path[0]).stem + ".html"
    generate_pyvis_graph(g, filename, args.use_progress_bar)


if __name__ == "__main__":
    main()



# Copyright 2023 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

import json
from pathlib import Path

import pytest

from surfactant.cmd.tui import TUI
from surfactant.plugin.manager import find_io_plugin, get_plugin_manager
from tests.cmd import common

testing_data = Path(Path(__file__).parent.parent, "data")


@pytest.mark.asyncio
async def test_generate(tmp_path):
    tui = TUI()
    async with tui.run_test() as pilot:
        # No good way to set files from here via the TUI - just set the path directly
        tui.generate_tab.specimen_context.input_path = (
            testing_data / "Windows_dll_test_no1"
        ).as_posix()
        tui.generate_tab.output_dir.input_path = tmp_path.as_posix()
        # Type in the filename
        await pilot.click(tui.generate_tab.output_name)
        await pilot.press(*"out.json")
        # Run it
        await pilot.click("#run")
    common.test_generate_result_no_install_prefix(
        (tmp_path / "out.json").as_posix(), (testing_data / "Windows_dll_test_no1").as_posix()
    )


@pytest.mark.asyncio
async def test_merge(tmp_path):
    tui = TUI()
    pm = get_plugin_manager()
    output_writer = find_io_plugin(pm, "cytrics", "write_sbom")
    input_reader = find_io_plugin(pm, "cytrics", "read_sbom")
    assert output_writer is not None
    assert input_reader is not None
    # Write the two SBOM's to files
    for i, sbom in enumerate((common.get_sbom1(), common.get_sbom2())):
        with open(tmp_path / f"sbom{i}.json", "w") as f:
            output_writer.write_sbom(sbom, f)
    async with tui.run_test() as pilot:
        # This is the only way I could figure out how to change tabs
        await pilot.press("right")
        # Add two merge paths (I don't know why the pauses are needed)
        await pilot.click("#add_input_path")
        await pilot.pause(0.5)
        await pilot.click("#add_input_path")
        await pilot.pause(0.5)
        # Set the merge paths
        for i, p in enumerate(tui.merge_tab.merge_paths.input_paths):
            p.path_selector.input_path = (tmp_path / f"sbom{i}.json").as_posix()
        tui.merge_tab.output_dir.input_path = tmp_path.as_posix()
        # Set the output name
        await pilot.click(tui.merge_tab.output_name)
        await pilot.press(*"test_out.json")
        # Run it
        await pilot.click(tui.merge_tab.btn)
    with open(tmp_path / "test_out.json") as f:
        merged_sbom = input_reader.read_sbom(f)
    common.test_simple_merge_method(common.get_sbom1(), common.get_sbom2(), merged_sbom)


__context_data = """
[
  {
    "archive": "/home/samples/helics.tar.gz",
    "extractPaths": ["/home/samples/helics"],
    "installPrefix": "/"
  },
  {
    "archive": "/home/samples/helics_plugin.tar.gz",
    "extractPaths": ["/home/samples/helics_plugin"],
    "installPrefix": "/"
  }
]
"""


@pytest.mark.asyncio
async def test_context_roundtrip(tmp_path):
    tui = TUI()
    with open(tmp_path / "test_input.json", "w") as f:
        f.write(__context_data)
    async with tui.run_test() as pilot:
        # Change to context tab
        await pilot.press(*("right", "right"))
        # Load an existing context file
        tui.context_tab.context_input.input_path = tmp_path.as_posix()
        tui.context_tab.context_name.value = "test_input.json"
        await pilot.click(tui.context_tab.load_btn)
        # Save it to a new file
        tui.context_tab.context_name.value = "test_output.json"
        await pilot.click(tui.context_tab.save_btn)
    # Compare the two JSON files
    with open(tmp_path / "test_input.json") as f:
        inp = json.load(f)
    with open(tmp_path / "test_output.json") as f:
        output = json.load(f)

    # There doesn't seem to be any better way of doing this
    def ordered(obj):
        if isinstance(obj, dict):
            return sorted(obj.items())
        return obj

    assert ordered(inp) == ordered(output)



# Copyright 2023 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

import json

from surfactant.sbomtypes import SBOM


def test_generate_result_no_install_prefix(output_path: str, extract_path: str):
    with open(output_path) as f:
        generated_sbom = json.load(f)

    assert len(generated_sbom["software"]) == 2

    expected_software_names = {"hello_world.exe", "testlib.dll"}
    actual_software_names = {software["fileName"][0] for software in generated_sbom["software"]}
    assert expected_software_names == actual_software_names

    expected_install_paths = {
        "hello_world.exe": extract_path + "/hello_world.exe",
        "testlib.dll": extract_path + "/testlib.dll",
    }
    for software in generated_sbom["software"]:
        assert software["installPath"][0] == expected_install_paths[software["fileName"][0]]

    uuids = {software["fileName"][0]: software["UUID"] for software in generated_sbom["software"]}
    assert len(generated_sbom["relationships"]) == 1
    assert generated_sbom["relationships"][0] == {
        "xUUID": uuids["hello_world.exe"],
        "yUUID": uuids["testlib.dll"],
        "relationship": "Uses",
    }


# Generate Sample SBOMs
sbom1_json = """{
  "software": [
    {
      "UUID": "dd6f7f6b-7c31-4a4a-afef-14678b9942bf",
      "fileName": [
        "helics.tar.gz"
      ]
    },
    {
      "UUID": "08526f02-8f08-485d-bfd1-ea16ce964fd2",
      "fileName": [
        "helics_binary"
      ],
      "installPath": [
        "/bin/helics_binary"
      ],
      "containerPath": [
        "dd6f7f6b-7c31-4a4a-afef-14678b9942bf/bin/helics_binary"
      ]
    },
    {
      "UUID": "a5db7e12-fe3d-490e-90b8-98a8bfaace09",
      "fileName": [
        "lib1.so"
      ],
      "installPath": [
        "/lib64/lib1.so"
      ],
      "containerPath": [
        "dd6f7f6b-7c31-4a4a-afef-14678b9942bf/lib64/lib1.so"
      ]
    }
  ],
  "relationships": [
    {
      "xUUID": "08526f02-8f08-485d-bfd1-ea16ce964fd2",
      "yUUID": "a5db7e12-fe3d-490e-90b8-98a8bfaace09",
      "relationship": "Uses"
    },
    {
      "xUUID": "dd6f7f6b-7c31-4a4a-afef-14678b9942bf",
      "yUUID": "08526f02-8f08-485d-bfd1-ea16ce964fd2",
      "relationship": "Contains"
    },
    {
      "xUUID": "dd6f7f6b-7c31-4a4a-afef-14678b9942bf",
      "yUUID": "a5db7e12-fe3d-490e-90b8-98a8bfaace09",
      "relationship": "Contains"
    }
  ]
}"""


def get_sbom1():
    return SBOM.from_json(sbom1_json)


sbom2_json = """{
        "software": [
            {
            "UUID": "625a07da-7eed-47b9-a0fa-47dcbf76574a",
            "name": null,
            "size": null,
            "fileName": [
                "helics_plugin.tar.gz"
            ]
            },
            {
            "UUID": "820d3ddc-14d7-4ce5-833c-beede8725366",
            "fileName": [
                "helics_plugin"
            ],
            "installPath": [
                "/bin/helics_plugin"
            ],
            "containerPath": [
                "625a07da-7eed-47b9-a0fa-47dcbf76574a/bin/helics_plugin"
            ]
            },
            {
            "UUID": "df81b6a7-f9df-42f1-85ee-86a8865fa5f1",
            "fileName": [
                "lib_plugin.so"
            ],
            "installPath": [
                "/lib64/lib_plugin.so"
            ],
            "containerPath": [
                "625a07da-7eed-47b9-a0fa-47dcbf76574a/lib64/lib_plugin.so"
            ]
            }
        ],
        "relationships": [
            {
            "xUUID": "820d3ddc-14d7-4ce5-833c-beede8725366",
            "yUUID": "df81b6a7-f9df-42f1-85ee-86a8865fa5f1",
            "relationship": "Uses"
            },
            {
            "xUUID": "625a07da-7eed-47b9-a0fa-47dcbf76574a",
            "yUUID": "820d3ddc-14d7-4ce5-833c-beede8725366",
            "relationship": "Contains"
            },
            {
            "xUUID": "625a07da-7eed-47b9-a0fa-47dcbf76574a",
            "yUUID": "df81b6a7-f9df-42f1-85ee-86a8865fa5f1",
            "relationship": "Contains"
            }
        ]
        }"""


def get_sbom2():
    return SBOM.from_json(sbom2_json)


def test_simple_merge_method(sbom1, sbom2, merged_sbom):
    # Capture each SBOM's original software entries
    orig_sw1, orig_sw2 = list(sbom1.software), list(sbom2.software)

    # Expect the merged list to be the union of the two originals
    expected_sw = orig_sw1 + orig_sw2
    assert sorted(merged_sbom.software, key=lambda x: x.UUID) == sorted(
        expected_sw, key=lambda x: x.UUID
    )

    # Verify graph edges union: use edges(keys=True) to pull relationship key
    def extract_edges(sbom):
        return set(sbom.graph.edges(keys=True))

    expected_edges = extract_edges(sbom1) | extract_edges(sbom2)
    assert extract_edges(merged_sbom) == expected_edges



from pathlib import Path
from typing import Optional

from loguru import logger

import surfactant.plugin
from surfactant.sbomtypes import SBOM

from . import visualization as vis


@surfactant.plugin.hookimpl
def write_sbom(sbom: SBOM, outfile) -> None:
    htmlName = Path(outfile.name).stem + ".html"

    logger.info(f"Generating visualization: {htmlName}")
    g = vis.generate_dependency_graph({"sbom": sbom, "sbomFileName": outfile.name})
    vis.generate_pyvis_graph(g, htmlName)

    logger.info("Writing CyTRICS format")
    outfile.write(sbom.to_json(indent=2))


@surfactant.plugin.hookimpl
def short_name() -> Optional[str]:
    return "sbomvis"



# Copyright 2023 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

import json
import os
import pathlib
import random
import string

import pytest

from surfactant.cmd.merge import merge
from surfactant.plugin.manager import get_plugin_manager
from surfactant.sbomtypes import SBOM
from tests.cmd import common


def get_config():
    return {
        "system": {
            "UUID": "6a0ee431-842f-4963-8867-ef0ef6998003",
            "name": "",
            "vendor": None,
            "captureStart": 1689186121,
            "captureEnd": 1689186146,
        }
    }


def get_sbom3():
    with open(
        pathlib.Path(__file__).parent / "../data/sample_sboms/helics_binaries_sbom.json",
        "r",
    ) as f:
        return SBOM.from_json(f.read())


def get_sbom4():
    with open(
        pathlib.Path(__file__).parent / "../data/sample_sboms/helics_libs_sbom.json", "r"
    ) as f:
        return SBOM.from_json(f.read())


# Test Functions
def test_simple_merge_method():
    sbom1 = common.get_sbom1()
    sbom2 = common.get_sbom2()

    # Merge in place
    merged_sbom = sbom1
    merged_sbom.merge(sbom2)

    common.test_simple_merge_method(common.get_sbom1(), common.get_sbom2(), merged_sbom)


@pytest.mark.skip(reason="No way of validating this test yet")
def test_merge_with_circular_dependency():
    sbom1 = common.get_sbom1()
    sbom2 = common.get_sbom2()
    circular_dependency_sbom = sbom1

    # inject a circular edge via the new graph API
    circular_dependency_sbom.create_relationship(
        "a5db7e12-fe3d-490e-90b8-98a8bfaace09",
        "dd6f7f6b-7c31-4a4a-afef-14678b9942bf",
        "Contains",
    )

    outfile_name = generate_filename("test_merge_with_circular_dependency")
    pm = get_plugin_manager()
    output_writer = pm.get_plugin("surfactant.output.cytrics_writer")
    input_sboms = [circular_dependency_sbom, sbom2]
    with open(outfile_name, "w") as sbom_outfile:
        merge(input_sboms, sbom_outfile, get_config(), output_writer)
    # TODO add validation checks here
    os.remove(os.path.abspath(outfile_name))


@pytest.mark.skip(reason="No way of properly validating this test yet")
def test_cmdline_merge():
    sbom3 = get_sbom3()
    sbom4 = get_sbom4()
    # Test simple merge of two sboms
    outfile_name = generate_filename("test_cmdline_merge")
    pm = get_plugin_manager()
    output_writer = pm.get_plugin("surfactant.output.cytrics_writer")
    config_file = get_config()
    input_sboms = [sbom3, sbom4]
    with open(outfile_name, "w") as sbom_outfile:
        merge(input_sboms, sbom_outfile, config_file, output_writer)

    # TODO add validation checks here
    with open(outfile_name, "r") as j:
        generated_sbom = json.loads(j.read())
    with open(pathlib.Path(__file__).parent / "../data/sample_sboms/helics_sbom.json", "r") as j:
        ground_truth_sbom = json.loads(j.read())
    os.remove(os.path.abspath(outfile_name))


def test_merge_with_add_system_true():
    sbom1 = common.get_sbom1()
    sbom2 = common.get_sbom2()
    outfile_name = generate_filename("test_merge_with_add_system_true")
    pm = get_plugin_manager()
    output_writer = pm.get_plugin("surfactant.output.cytrics_writer")
    input_sboms = [sbom1, sbom2]
    config_file = get_config()
    config_file["system"]["UUID"] = "6a0ee431-842f-4963-8867-ef0ef6998003"
    with open(outfile_name, "w") as sbom_outfile:
        merge(input_sboms, sbom_outfile, config_file, output_writer, add_system=True)

    with open(outfile_name, "r") as j:
        generated_sbom = json.loads(j.read())
    assert generated_sbom["systems"]
    assert generated_sbom["systems"][0]["UUID"] == config_file["system"]["UUID"]

    os.remove(os.path.abspath(outfile_name))


def test_merge_with_add_system_false():
    sbom1 = common.get_sbom1()
    sbom2 = common.get_sbom2()
    outfile_name = generate_filename("test_merge_with_add_system_false")
    pm = get_plugin_manager()
    output_writer = pm.get_plugin("surfactant.output.cytrics_writer")
    input_sboms = [sbom1, sbom2]
    config_file = get_config()
    config_file["system"]["UUID"] = "6a0ee431-842f-4963-8867-ef0ef6998003"
    with open(outfile_name, "w") as sbom_outfile:
        merge(input_sboms, sbom_outfile, config_file, output_writer, add_system=False)

    with open(outfile_name, "r") as j:
        generated_sbom = json.loads(j.read())
    assert not generated_sbom["systems"]

    os.remove(os.path.abspath(outfile_name))


def test_merge_with_custom_system_relationship():
    sbom1 = common.get_sbom1()
    sbom2 = common.get_sbom2()
    outfile_name = generate_filename("test_merge_with_custom_system_relationship")
    pm = get_plugin_manager()
    output_writer = pm.get_plugin("surfactant.output.cytrics_writer")
    input_sboms = [sbom1, sbom2]
    config_file = get_config()
    config_file["systemRelationship"] = "DependsOn"
    with open(outfile_name, "w") as sbom_outfile:
        merge(input_sboms, sbom_outfile, config_file, output_writer, add_system=True)

    with open(outfile_name, "r") as j:
        generated_sbom = json.loads(j.read())
    for relationship in generated_sbom["relationships"]:
        if relationship["xUUID"] == config_file["system"]["UUID"]:
            assert relationship["relationship"] == "DependsOn"

    os.remove(os.path.abspath(outfile_name))


def test_merge_with_specified_system_uuid():
    sbom1 = common.get_sbom1()
    sbom2 = common.get_sbom2()
    outfile_name = generate_filename("test_merge_with_specified_system_uuid")
    pm = get_plugin_manager()
    output_writer = pm.get_plugin("surfactant.output.cytrics_writer")
    input_sboms = [sbom1, sbom2]
    config_file = get_config()
    system_uuid = "123e4567-e89b-12d3-a456-426614174000"
    with open(outfile_name, "w") as sbom_outfile:
        merge(
            input_sboms,
            sbom_outfile,
            config_file,
            output_writer,
            add_system=True,
            system_uuid=system_uuid,
        )

    with open(outfile_name, "r") as j:
        generated_sbom = json.loads(j.read())
    assert any(system["UUID"] == system_uuid for system in generated_sbom["systems"])

    os.remove(os.path.abspath(outfile_name))


def test_prevent_orphaned_system_uuid():
    sbom1 = common.get_sbom1()
    sbom2 = common.get_sbom2()
    outfile_name = generate_filename("test_prevent_orphaned_system_uuid")
    pm = get_plugin_manager()
    output_writer = pm.get_plugin("surfactant.output.cytrics_writer")
    input_sboms = [sbom1, sbom2]
    config_file = get_config()
    # Get rid of the system UUID field from the config file
    # This will make it try to generate a random UUID, but won't add it since add_system is False
    del config_file["system"]["UUID"]
    with open(outfile_name, "w") as sbom_outfile:
        merge(input_sboms, sbom_outfile, config_file, output_writer, add_system=False)

    with open(outfile_name, "r") as j:
        generated_sbom = json.loads(j.read())
    assert not generated_sbom["systems"]

    os.remove(os.path.abspath(outfile_name))


def test_add_random_system_uuid():
    sbom1 = common.get_sbom1()
    sbom2 = common.get_sbom2()
    outfile_name = generate_filename("test_prevent_orphaned_system_uuid")
    pm = get_plugin_manager()
    output_writer = pm.get_plugin("surfactant.output.cytrics_writer")
    input_sboms = [sbom1, sbom2]
    config_file = get_config()
    # Get rid of the system UUID field from the config file
    # This will make it try to generate a random UUID, but won't add it since add_system is False
    original_config_system_UUID = config_file["system"]["UUID"]
    del config_file["system"]["UUID"]
    with open(outfile_name, "w") as sbom_outfile:
        merge(input_sboms, sbom_outfile, config_file, output_writer, add_system=True)

    with open(outfile_name, "r") as j:
        generated_sbom = json.loads(j.read())
    assert generated_sbom["systems"]
    # Check that the UUID of the generated system is actually random
    assert generated_sbom["systems"][0]["UUID"] != original_config_system_UUID

    os.remove(os.path.abspath(outfile_name))


def generate_filename(name, ext=".json"):
    res = "".join(random.choices(string.ascii_uppercase + string.digits, k=7))
    return str(name + "_" + res + ext)
